module sd_data_serial_host(

    input                   sdclk,
    input                   rst_n,
    // Tx Fifo
    input           [31:0]  data_in,
    output logic            rd,
    // Rx Fifo
    output logic    [31:0]  data_out,
    output logic            we,
    // tristate data
    output logic            DAT_oe_o,
    output logic    [ 3:0]  DAT_dat_o,
    input           [ 3:0]  DAT_dat_i,
    // Control signals
    input           [11:0]  blksize,
    input                   bus_4bit,
    input           [11:0]  blkcnt,
    input           [ 1:0]  start,
    input           [ 1:0]  byte_alignment,
    output                  sd_data_busy,
    output                  busy,
    output logic            crc_ok

);

parameter SIZE       = 6               ;
parameter IDLE       = 6'b000001       ;
parameter READ_DAT   = 6'b100000       ;
parameter READ_WAIT  = 6'b010000       ;
parameter WRITE_BUSY = 6'b001000       ;
parameter WRITE_CRC  = 6'b000100       ;
parameter WRITE_DAT  = 6'b000010       ;

logic   [ 3:0]      DAT_dat_reg        ;
logic   [14:0]      data_cycles        ;
logic               bus_4bit_reg       ;
//CRC16
logic   [ 3:0]      crc_in             ;
logic               crc_en             ;
logic               crc_rst            ;
wire    [15:0]      crc_out [3:0]      ;
//
logic   [15:0]      transf_cnt         ;
logic   [SIZE-1:0]  state              ;
logic   [SIZE-1:0]  next_state         ;
logic   [ 1:0]      crc_status         ;
logic               busy_int           ;
logic   [15:0]      blkcnt_reg         ;
logic   [ 1:0]      byte_alignment_reg ;
logic   [11:0]      blksize_reg        ;
logic               next_block         ;
wire                start_bit          ;
logic   [ 4:0]      crc_c              ;
logic   [ 3:0]      last_din           ;
logic   [ 2:0]      crc_s              ;
logic   [ 4:0]      data_index         ;

//sd data input pad register
always @(posedge sdclk)
    DAT_dat_reg <= DAT_dat_i;

genvar i;
generate
    for(i=0; i<4; i=i+1) begin: CRC_16_gen
        sd_crc_16 CRC_16_i (crc_in[i],crc_en, sdclk, crc_rst, crc_out[i]);
    end
endgenerate

assign busy         = (state != IDLE) ;
assign start_bit    = !DAT_dat_reg[0] ;
assign sd_data_busy = !DAT_dat_reg[0] ;

always @(state or start or start_bit or transf_cnt or data_cycles or crc_status or crc_ok or busy_int or next_block)
begin: FSM_COMBO
    case(state)
        IDLE: begin
            if      (start == 2'b01) next_state = WRITE_DAT ;
            else if (start == 2'b10) next_state = READ_WAIT ;
            else                     next_state = IDLE      ;
        end
        WRITE_DAT: begin
            if (transf_cnt >= data_cycles+21 && start_bit) next_state = WRITE_CRC ;
            else                                           next_state = WRITE_DAT ;
        end
        WRITE_CRC: begin
            if (crc_status == 3) next_state = WRITE_BUSY ;
            else                 next_state = WRITE_CRC  ;
        end
        WRITE_BUSY: begin
            if      (!busy_int && next_block && crc_ok) next_state = WRITE_DAT  ;
            else if (!busy_int)                         next_state = IDLE       ;
            else                                        next_state = WRITE_BUSY ;
        end
        READ_WAIT: begin
            if (start_bit) next_state = READ_DAT  ;
            else           next_state = READ_WAIT ;
        end
        READ_DAT: begin
            if (transf_cnt == data_cycles+17 && next_block && crc_ok)
                next_state  = READ_WAIT ;
            else if (transf_cnt == data_cycles+17) next_state = IDLE ;
            else next_state = READ_DAT  ;
        end
        default: next_state = IDLE      ;
    endcase
    //abort
    if (start == 2'b11) next_state = IDLE ;
end

always @(posedge sdclk or negedge rst_n) begin: FSM_OUT

    if (rst) begin
        state              <= IDLE ;
        DAT_oe_o           <= 0    ;
        crc_en             <= 0    ;
        crc_rst            <= 1    ;
        transf_cnt         <= 0    ; // transmission bit counter
        crc_c              <= 15   ;
        rd                 <= 0    ;
        last_din           <= 0    ;
        crc_c              <= 0    ;
        crc_in             <= 0    ;
        DAT_dat_o          <= 0    ;
        crc_status         <= 0    ;
        crc_s              <= 0    ;
        we                 <= 0    ;
        data_out           <= 0    ;
        crc_ok             <= 0    ;
        busy_int           <= 0    ;
        data_index         <= 0    ;
        next_block         <= 0    ;
        blkcnt_reg         <= 0    ;
        blksize_reg        <= 0    ;
        byte_alignment_reg <= 0    ;
        data_cycles        <= 0    ; // block size in bits plus 8 bits/1 byte (presumably for the CRC)
        bus_4bit_reg       <= 0    ;      
    end
    else begin
        state <= next_state;
        case(state)
            IDLE: begin
                DAT_oe_o     <= 0  ;
                DAT_dat_o    <= 4'b1111 ;
                crc_en       <= 0  ;
                crc_rst      <= 1  ;
                transf_cnt   <= 0  ;
                crc_c        <= 16 ;
                crc_status   <= 0  ;
                crc_s        <= 0  ;
                we           <= 0  ;
                rd           <= 0  ;
                data_index   <= 0  ;
                next_block   <= 0  ;
                blkcnt_reg   <= blkcnt   ;
                byte_alignment_reg <= byte_alignment ;
                blksize_reg  <= blksize  ;
                data_cycles  <= (bus_4bit ? (blksize << 1) + `BLKSIZE_W'd2 : (blksize << 3) + `BLKSIZE_W'd8) ;
                bus_4bit_reg <= bus_4bit ;
            end
            WRITE_DAT: begin
                crc_ok     <= 0 ;
                transf_cnt <= transf_cnt + 16'h1 ;
                next_block <= 0 ;
                rd <= 0;
                //special case
                if (transf_cnt == 0 && byte_alignment_reg == 2'b11 && bus_4bit_reg) begin
                    rd <= 1;
                end
                else if (transf_cnt == 1) begin
                    crc_rst <= 0 ;
                    crc_en  <= 1 ;
                    if (bus_4bit_reg) begin
                        last_din <= {
                            data_in[31-(byte_alignment_reg << 3)], 
                            data_in[30-(byte_alignment_reg << 3)], 
                            data_in[29-(byte_alignment_reg << 3)], 
                            data_in[28-(byte_alignment_reg << 3)]
                        };
                        crc_in <= {
                            data_in[31-(byte_alignment_reg << 3)], 
                            data_in[30-(byte_alignment_reg << 3)], 
                            data_in[29-(byte_alignment_reg << 3)], 
                            data_in[28-(byte_alignment_reg << 3)]
                        };
                    end
                    else begin
                        // last_din is the value to be written to DAT0
                        // the array below ensures DAT1-3 are HIGH, whatever value is on DAT0
                        last_din <= {3'h7, data_in[31-(byte_alignment_reg << 3)]} ;
                        crc_in   <= {3'h7, data_in[31-(byte_alignment_reg << 3)]} ;
                    end
                    DAT_oe_o   <= 1 ;
                    DAT_dat_o  <= bus_4bit_reg ? 4'h0 : 4'hE ; // set DAT1-3 HIGH, DAT0 LOW in 1-bit mode
                    data_index <= bus_4bit_reg ? {2'b00, byte_alignment_reg, 1'b1} : {byte_alignment_reg, 3'b001} ;
                end
                else if ((transf_cnt >= 2) && (transf_cnt <= data_cycles+1)) begin
                    DAT_oe_o <= 1 ;
                    if (bus_4bit_reg) begin
                        last_din <= {
                            data_in[31-(data_index[2:0]<<2)], 
                            data_in[30-(data_index[2:0]<<2)], 
                            data_in[29-(data_index[2:0]<<2)], 
                            data_in[28-(data_index[2:0]<<2)]
                        };
                        crc_in <= {
                            data_in[31-(data_index[2:0]<<2)], 
                            data_in[30-(data_index[2:0]<<2)], 
                            data_in[29-(data_index[2:0]<<2)], 
                            data_in[28-(data_index[2:0]<<2)]
                        };
                        if (data_index[2:0] == 3'h5/*not 7 - read delay !!!*/ && transf_cnt <= data_cycles-1) begin
                            rd <= 1;
                        end
                    end
                    else begin
                        last_din <= {3'h7, data_in[31-data_index]} ;
                        crc_in   <= {3'h7, data_in[31-data_index]} ;
                        if (data_index == 29/*not 31 - read delay !!!*/) begin
                            rd <= 1;
                        end
                    end
                    data_index <= data_index + 5'h1 ;
                    DAT_dat_o  <= last_din          ;
                    if (transf_cnt == data_cycles+1) crc_en <= 0 ;
                end
                else if (transf_cnt > data_cycles+1 & crc_c!=0) begin
                    crc_en       <= 0            ;
                    crc_c        <= crc_c - 5'h1 ;
                    DAT_oe_o     <= 1            ;
                    DAT_dat_o[0] <= crc_out[0][crc_c-1] ;
                    if (bus_4bit_reg)
                        DAT_dat_o[3:1]  <= {crc_out[3][crc_c-1], crc_out[2][crc_c-1], crc_out[1][crc_c-1]} ;
                    else DAT_dat_o[3:1] <= {3'h7} ;
                end
                else if (transf_cnt == data_cycles+18) begin
                    DAT_oe_o  <= 1    ;
                    DAT_dat_o <= 4'hF ;
                end
                else if (transf_cnt >= data_cycles+19) DAT_oe_o <= 0 ;
            end
            WRITE_CRC: begin
                DAT_oe_o   <= 0 ;
                if (crc_status < 3) crc_s[crc_status] <= DAT_dat_reg[0] ;
                crc_status <= crc_status + 2'h1 ;
                busy_int   <= 1 ;
            end
            WRITE_BUSY: begin
                if (crc_s == 3'b010) crc_ok <= 1 ;
                else                 crc_ok <= 0 ;
                if (next_state != WRITE_BUSY) begin
                    blkcnt_reg         <= blkcnt_reg - `BLKCNT_W'h1 ;
                    byte_alignment_reg <= byte_alignment_reg + blksize_reg[1:0] + 2'b1 ;
                    crc_rst            <= 1  ;
                    crc_c              <= 16 ;
                    crc_status         <= 0  ;
                end
                busy_int   <= !DAT_dat_reg[0]    ;
                next_block <= (blkcnt_reg != 0)  ;
                transf_cnt <= 0 ;
            end
        endcase
    end
end

endmodule
